/**
 * $OpenXM: OpenXM/src/OpenMath/ORG/openxm/tam/OpenXM.java,v 1.10 2002/10/27 10:39:32 takayama Exp $
 */
package ORG.openxm.tam;

import java.io.*;
import java.net.*;


/*&ja
    OpenXM ФȤ³Ԥʤ饹. 
    饤¦Ѥ.     
    ³륵˰Ĥ OpenXM 饹ɬ. 
*/
/**
 * OpenXM is a class to connect to OpenXM servers,
 * which are compliant to OpenXM RFC 100.
 * There is one-to-one correspondence between the instances
 * of the class OpenXM and the OpenXM servers.
 */
public class OpenXM{
  private OpenXMstream control = null, stream = null;
  final protected boolean debug = false;

  /*&ja
   * OpenXM ФȤ³ TCP/IP åȤѤƹԤʤ.
   * ޥ̾ host Υݡֹ CtrlPort ˥ȥ,
   * ݡֹ StreamPort ˥ǡѤ³Ԥʤ.
   */
  /**
   * Connect to an OpenXM server via TCP/IP socket.
   * @param host  a machine name of the OpenXM server.
   * @param CtrlPort  the control port number of the OpenXM server.
   * @param StreamPort  the data port number of the OpenXM server.
   * As to details on the notion of control port and data port, see 
   *    Design and Implementation of OpenXM client server model and
   *        common mathematical object format (OpenXM-RFC 100,
   *                                            proposed standard) 
   *   @see <a href="http://www.openxm.org">OpenXM</a>
   */
  public OpenXM(String host,int CtrlPort,int StreamPort) throws IOException{
    control = new OpenXMstream(host,CtrlPort);

    try{
      Thread.sleep(100); // We need a few wait for starting up server.
    }catch(InterruptedException e){
      System.err.println(e.getMessage());
    }

    stream = new OpenXMstream(host,StreamPort);

    control.sendByteOrder();
    stream.sendByteOrder();
  }

  /*&ja
   * ޥ command Ω夲
   * OpenXM ФȤ³ TCP/IP åȤѤƹԤʤ.
   * ޥ̾ host Υݡֹ CtrlPort ˥ȥ,
   * ݡֹ StreamPort ˥ǡѤ³Ԥʤ.
   */
  /**
   * First, execute a command, which is usually an OpenXM server,
   * and next try to connect to the OpenXM server via TCP/IP.
   * @param command  a command.
   * @param host  a machine name of the OpenXM server.
   * @param CtrlPort  the control port number of the OpenXM server.
   * @param StreamPort  the data port number of the OpenXM server.
   */
  public OpenXM(String command,String host,int CtrlPort,int StreamPort)
       throws IOException{
    Runtime.getRuntime().exec(command);
    control = new OpenXMstream(host,CtrlPort);

    try{
      Thread.sleep(100); // We need a few wait for starting up server.
    }catch(InterruptedException e){
      System.err.println(e.getMessage());
    }

    stream = new OpenXMstream(host,StreamPort);

    control.sendByteOrder();
    stream.sendByteOrder();
  }

  public OpenXM(String host,int CtrlPort,int StreamPort,String ox_server,int oxd_port,String pass)
	   throws IOException{
		 // pass may be a null string.
    control = new OpenXMstream(host,CtrlPort,pass);
    stream = new OpenXMstream(host,StreamPort,pass);
    System.err.println("Listening...");
    System.err.println("Launch ox server with the reverse option, e.g., ox -ox ox_asir -reverse");
    // Launch ox_server by the oxd daemon (oxd_port).
    // BUG: It has not yet been implemented.
    control.OpenXMstreamAccept();
    System.err.println("Accepted the control port.");
    stream.OpenXMstreamAccept();
    System.err.println("Accepted the data port.");


    control.sendByteOrder();
    stream.sendByteOrder();
  }

public OpenXM(String ox_server) throws IOException {
	int oxdPort = 8089;
	String host = "localhost";
	oxdStream oxd = new oxdStream(oxdPort);
	int cport,dport;
	cport = oxd.startPhase1();
	dport = cport+1;
	try {
		control = new OpenXMstream(host,cport,"");
		stream = new OpenXMstream(host,dport,"");
	}catch( IOException e) {
		System.err.println("Could not open ports for client.");
		oxd.write("<bye/>");
	}

	System.err.println("Listenning...");

	oxd.startPhase2(ox_server,cport);
	
    control.OpenXMstreamAccept();
    System.err.println("Accepted the control port.");
    stream.OpenXMstreamAccept();
    System.err.println("Accepted the data port.");


    control.sendByteOrder();
    stream.sendByteOrder();
}


  /*&ja
   * Фη׻ǤԤʤ. ߤ̤.
   */
  /**
   * Resetting the engine process.  It has not yet been implemented.
   */
  public synchronized void resetConnection(){
    debug("control: stopping computer process...");
    debug("control: sending SYNC BALL.");
  }

  /*&ja
   * OpenXM åǡȥ꡼.
   * Υ᥽åɤϥåΥܥǥʬǤ褤.
   * إåʬϼưղä.
   */
  /**
   * Send an OpenXM message object. 
   * @param object a message. For example, oxm.send(new CMO_STRING("Hello"))
   * sends a string "Hello" to the OpenXM server oxm in the CMO_STRING data
   * encoding.
   */
  public void send(OXbody object) throws IOException,MathcapViolation{
    stream.send(object);
  }

  /*&ja
   * ǡȥ꡼फ OpenXM å.
   */
  /**
   * Receive an OpenXM message.
   */
  public OXmessage receive() throws IOException{
    return stream.receive();
  }

  /*&ja
   * ǡȥ꡼ MathCap  mathcap ꤹ.
   * ʸ, 륪֥Ȥ mathcap ˹äƤ뤫ɤ
   * å. ºݤ˥å뤫ɤ
   * OXbody 饹饹μˤ.
   * mathcap ȿ֥ȤȤˤ,
   * ʸ MathcapViolation ȯ뤳ȤԤ.
   */
  /**
   * Set the mathcap.
   * If one tries to send an object which is prohibited to send by the mathcap,
   * the mathcapViolation exception is thrown.
   */
  public void setMathCap(CMO_MATHCAP mathcap){
    stream.setMathCap(mathcap);
  }

  private final void debug(String str){
    if(debug){
      System.err.println(str);
    }
  }

  public static void main(String[] argv){
    String hostname = "localhost";
    int ControlPort = 1200, DataPort = 1300;
    Runnable process = null;
    Thread thread;
    OpenXM ox;

    for(int i=0;i<argv.length;i++){
      if(argv[i].equals("-h")){
        System.err.println("");
        System.exit(0);
      }else if(argv[i].equals("-host")){
        hostname = argv[++i];
      }else if(argv[i].equals("-data")){
        DataPort = Integer.valueOf(argv[++i]).intValue();
      }else if(argv[i].equals("-control")){
        ControlPort = Integer.valueOf(argv[++i]).intValue();
      }else{
        System.err.println("unknown option :"+ argv[i]);
        System.exit(1);
      }
    }

    try{
      ox = new OpenXM(hostname,ControlPort,DataPort);

      thread = new Thread(process);
      thread.start();
    }catch(UnknownHostException e){
      System.err.println("host unknown.");
      System.err.println(e.getMessage());
      return;
    }catch(IOException e){
      System.err.println("connection failed.");
      System.err.println("IOException occuer !!");
      System.err.println(e.getMessage());
      return;
    }

    try{
      //¦ʸޤ
      ox.send(new SM(SM.SM_mathcap));
      ox.send(new SM(SM.SM_popString));

      //ox.send(new CMO_STRING("print(\"Hello world!!\");\n"));
      //ox.send(new SM(SM.SM_executeStringByLocalParser));

      //ox.send(new CMO_STRING("def sub(B,A){return B-A;}"));
      //ox.send(new SM(SM.SM_executeStringByLocalParser));
      //ox.send(new SM(SM.SM_popString));

      //ox.send(new CMO_STRING("diff((x+2*y)^2,x);\0 1+2;"));
      //ox.send(new SM(SM.SM_executeStringByLocalParser));
      //ox.send(new SM(SM.SM_popString));

      //ox.send(new CMO_ZZ("1"));
      ox.send(new CMO_ZZ("-2"));
      //ox.send(new CMO_INT32(2));
      //ox.send(new CMO_STRING("sub"));
      //ox.send(new SM(SM.SM_executeFunction));
      ox.send(new SM(SM.SM_popCMO));

      { int[] array = {1,2};
      ox.send(new CMO_MONOMIAL32(array,new CMO_ZZ("-2")));
      }
      ox.send(new SM(SM.SM_popCMO));

      ox.send(new SM(SM.SM_popString));
      //ox.send(new SM(SM.SM_popString));

      ox.send(new SM(SM.SM_control_kill));

      //os.flush();

      //¦줿ʸޤ
      while(true){
	ox.receive();
      }

    }catch(Exception e){
      e.printStackTrace();
    }

    System.err.println("breaking...");

    try{
      // close connection
      ox.stream.close();
      ox.control.close();
    }catch(IOException e){
      e.printStackTrace();
    }
  }
}
