with Standard_Complex_Polynomials;       use Standard_Complex_Polynomials;
with Standard_Complex_Poly_Matrices;
with Brackets;                           use Brackets;
with Bracket_Monomials;                  use Bracket_Monomials;
with Bracket_Systems;                    use Bracket_Systems;

package Symbolic_Minor_Equations is

-- DESCRIPTION :
--   This package generates the equations that arise from the intersection
--   conditions along the deformations determined by the Pieri trees.
--   The equations are symbolic: the coefficients are brackets.

-- LOCALIZATION PATTERNS :

  function Schubert_Pattern ( n : natural; b1,b2 : Bracket )
                            return Standard_Complex_Poly_Matrices.Matrix;

  -- DESCRIPTON :
  --   Returns the representation of the pattern of the p-plane that satisfies
  --   the Schubert conditions as a polynomial matrix.
  --   This definition is used in the original Pieri implementation.

  function Localization_Pattern ( n : natural; top,bottom : Bracket )
                                return Standard_Complex_Poly_Matrices.Matrix;

  -- DESCRIPTION :
  --   Returns the matrix of indeterminates for the top and bottom pivots.
  --   The dimension of the working space equals n.
  --   This definition is used in the second Pieri implementation.

-- SYMBOLIC REPRESENTATIONS OF THE EQUATIONS :

  function Maximal_Minors ( n,m : natural ) return Bracket_Monomial;

  -- DESCRIPTION :
  --   Generates all maximal minors of an n-by-m matrix, n >= m.
  --   The collection of all these m-by-m minors is represented as the
  --   product of m-brackets.  Every bracket determines a selection of
  --   m rows from the n-by-m matrix.

  function Minor_Equations
            ( m,d : natural; bm : Bracket_Monomial ) return Bracket_System;

  -- DESCRIPTION :
  --   Returns the bracket polynomials that arise from expanding the
  --   above maximal m-minors into blocks of respective sizes m-d and d.
  --   Equation 0 in the result is the generic bracket representation of
  --   the Laplace expansion.

  function Expanded_Minor ( m : Standard_Complex_Poly_Matrices.Matrix;
                            b : Bracket ) return Poly;

  -- DESCRIPTION :
  --   The minor that selects rows from m, according to b is expanded.

  function Extend_Zero_Lifting ( p : Poly ) return Poly;

  -- DESCRIPTION :
  --   Extends every term with a new variable t^0.

end Symbolic_Minor_Equations;
